/* Copyright (c) 2009, Cornell University.  All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer
       in the documentation and/or other materials provided with the
       distribution.
     * Neither the name of the University nor the names of its
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#ifndef ALLY_H
#define ALLY_H

#include <stddef.h>


/* Communication with peers. */

typedef struct ally_peers_t ally_peers_t;
typedef int (*ally_peers_send_t) (ally_peers_t *peers, size_t peer,
                                  size_t size, const char *message);
typedef void (*ally_peers_done_t) (ally_peers_t *peers);

struct ally_peers_t
{
  size_t size;
  ally_peers_send_t send;
  ally_peers_done_t done;
};


/* Distributed cryptosystems. */

typedef struct ally_crypt_t ally_crypt_t;
typedef struct ally_crypt_public_t ally_crypt_public_t;
typedef struct ally_crypt_private_t ally_crypt_private_t;

typedef int (*ally_crypt_deal_t) (ally_crypt_t *crypt);
typedef ally_crypt_public_t *(*ally_crypt_get_public_t) (ally_crypt_t *crypt);
typedef ally_crypt_private_t *(*ally_crypt_get_private_t) (ally_crypt_t
                                                           *crypt,
                                                           size_t peer);
typedef void (*ally_crypt_done_t) (ally_crypt_t *crypt);

typedef int (*ally_crypt_public_combine_t) (ally_crypt_public_t *pub,
                                            size_t num, const char **partials,
                                            char *signature);
typedef int (*ally_crypt_public_verify_t) (ally_crypt_public_t *pub,
                                           char *signature);
typedef void (*ally_crypt_public_done_t) (ally_crypt_public_t *pub);

typedef int (*ally_crypt_private_sign_t) (ally_crypt_private_t *priv,
                                          size_t size, const char *message,
                                          char *partial);
typedef void (*ally_crypt_private_done_t) (ally_crypt_private_t *priv);

struct ally_crypt_t
{
  size_t size;
  ally_crypt_deal_t deal;
  ally_crypt_get_public_t get_public;
  ally_crypt_get_private_t get_private;
  ally_crypt_done_t done;
};

struct ally_crypt_public_t
{
  size_t signature_size;
  ally_crypt_public_combine_t combine;
  ally_crypt_public_verify_t verify;
  ally_crypt_public_done_t done;
};

struct ally_crypt_private_t
{
  size_t signature_size;
  ally_crypt_private_sign_t sign;
  ally_crypt_private_done_t done;
};


/* Agreement protocols. */

typedef struct ally_agree_t ally_agree_t;
typedef int (*ally_agree_message_t) (ally_agree_t *agree, size_t size,
                                     const char *message);
typedef int (*ally_agree_propose_t) (ally_agree_t *agree, size_t size,
                                     const char *message);
typedef int (*ally_agree_learn_t) (ally_agree_t *agree, size_t size,
                                   const char *message);
typedef void (*ally_agree_done_t) (ally_agree_t *agree);

struct ally_agree_t
{
  ally_agree_message_t message;
  ally_agree_propose_t propose;
  ally_agree_done_t done;
};


/* Shoup distributed cryptosystem. */

ally_crypt_t *ally_crypt_shoup (size_t l, size_t t, size_t k, size_t bits);


/* Fast Asynchronous Byzantine Paxos agreement protocol. */

typedef enum ally_agree_fab_paxos_role_t ally_agree_fab_paxos_role_t;

enum ally_agree_fab_paxos_role_t
{
  ALLY_AGREE_FAB_PAXOS_ROLE_PROPOSER,
  ALLY_AGREE_FAB_PAXOS_ROLE_ACCEPTOR,
  ALLY_AGREE_FAB_PAXOS_ROLE_LEARNER
};

ally_agree_t *ally_agree_fab_paxos (ally_peers_t *proposers,
                                    ally_peers_t *acceptors,
                                    ally_peers_t *learners,
                                    ally_agree_fab_paxos_role_t role,
                                    ally_agree_learn_t learn, size_t self);


#endif /* not ALLY_H */
