/* Copyright (c) 2009, Cornell University.  All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer
       in the documentation and/or other materials provided with the
       distribution.
     * Neither the name of the University nor the names of its
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include "config.h"
#include <ally.h>
#include <stdlib.h>
#include <time.h>
#include <gmp.h>

static const unsigned long TRIALS = 5;

typedef struct ally_crypt_shoup_t ally_crypt_shoup_t;
typedef struct ally_crypt_public_shoup_t ally_crypt_public_shoup_t;
typedef struct ally_crypt_private_shoup_t ally_crypt_private_shoup_t;

struct ally_crypt_shoup_t
{
  ally_crypt_t super;
  gmp_randstate_t rnd;
  size_t bits;
  size_t t;
  size_t k;
  mpz_t n;
  mpz_t m;
  mpz_t p;
  mpz_t q;
  mpz_t e;
  mpz_t d;
  mpz_t delta;
  mpz_t invdelta;
  mpz_t a[];
};

struct ally_crypt_public_shoup_t
{
  ally_crypt_public_t super;
};

struct ally_crypt_private_shoup_t
{
  ally_crypt_private_t super;
  mpz_t s;
};

static int
deal (ally_crypt_shoup_t * crypt)
{
  size_t i;
  mpz_t tmp;

  /* Choose an RSA modulus based on strong primes p and q. */
  do
    {
      mpz_urandomb (crypt->p, crypt->rnd, crypt->bits - 2);
      mpz_urandomb (crypt->q, crypt->rnd, crypt->bits - 2);
      mpz_setbit (crypt->p, crypt->bits - 1);
      mpz_setbit (crypt->q, crypt->bits - 1);
      mpz_nextprime (crypt->p, crypt->p);
      mpz_nextprime (crypt->q, crypt->q);
      mpz_mul (crypt->m, crypt->p, crypt->q);
      mpz_mul_2exp (crypt->p, crypt->p, 1);
      mpz_mul_2exp (crypt->q, crypt->q, 1);
      mpz_add_ui (crypt->p, crypt->p, 1);
      mpz_add_ui (crypt->q, crypt->q, 1);
      mpz_mul (crypt->n, crypt->p, crypt->q);
    }
  while (!mpz_probab_prime_p (crypt->p, TRIALS)
         && !mpz_probab_prime_p (crypt->q, TRIALS));

  /* Find and split the private key. */
  mpz_init (tmp);
  mpz_invert (crypt->p, crypt->e, crypt->m); /* FIXME: Check for (and bail on) failure. */
  mpz_set (crypt->a[0], crypt->p);
  mpz_sub_ui (crypt->m, crypt->m, 1);
  for (i = 1; i < crypt->super.size; ++i)
    mpz_urandomm (crypt->a[i], crypt->rnd, crypt->m);
  mpz_add_ui (crypt->m, crypt->m, 1);

  /* Invert delta for use in finding private keys. */
  mpz_invert (crypt->invdelta, crypt->delta, crypt->m);
  return -1;
}

static int
combine (ally_crypt_public_shoup_t *pub, size_t num, const char **partials, char *signature)
{
  /* FIXME: implement */
  return -1;
}

static int
verify (ally_crypt_public_shoup_t *pub, char *signature)
{
  /* FIXME: implement */
  return -1;
}

static int
sign (ally_crypt_private_shoup_t *priv, size_t size, const char *message, char *partial)
{
  /* FIXME: implement */
  return -1;
}

static void
done (ally_crypt_shoup_t *crypt)
{
  free (crypt);
}

static void
public_done (ally_crypt_public_shoup_t *pub)
{
  free (pub);
}

static void
private_done (ally_crypt_private_shoup_t *priv)
{
  free (priv);
}

static ally_crypt_public_t *
get_public (ally_crypt_shoup_t *crypt)
{
  ally_crypt_public_shoup_t *pub = malloc (sizeof *pub);
  if (pub)
    {
      pub->super.signature_size = 0; /* FIXME: based on bits */
      pub->super.combine = (ally_crypt_public_combine_t) combine;
      pub->super.verify = (ally_crypt_public_verify_t) verify;
      pub->super.done = (ally_crypt_public_done_t) public_done;
      /* FIXME: initialize from dealt values */
    }
  return (ally_crypt_public_t *) pub;
}

static ally_crypt_private_t *
get_private (ally_crypt_shoup_t *crypt, size_t peer)
{
  ally_crypt_private_shoup_t *priv = malloc (sizeof *priv);
  if (priv)
    {
      size_t i = crypt->super.size;
      priv->super.signature_size = 0; /* FIXME: based on bits */
      priv->super.sign = (ally_crypt_private_sign_t) sign;
      priv->super.done = (ally_crypt_private_done_t) private_done;
      mpz_init (priv->s);
      while (i > 0)
        {
          mpz_mul_ui (priv->s, priv->s, peer + 1);
	  mpz_add (priv->s, priv->s, crypt->a[--i]);
	  mpz_mod (priv->s, priv->s, crypt->m);
	}
      mpz_mul (priv->s, priv->s, crypt->invdelta);
      mpz_mod (priv->s, priv->s, crypt->m);
    }
  return (ally_crypt_private_t *) priv;
}

ally_crypt_t *
ally_crypt_shoup (size_t l, size_t t, size_t k, size_t bits)
{
  ally_crypt_shoup_t *crypt = malloc (sizeof *crypt + l * sizeof crypt->a[0]);
  if (crypt)
    {
      size_t i;
      crypt->super.size = l;
      crypt->super.get_public = (ally_crypt_get_public_t) get_public;
      crypt->super.get_private = (ally_crypt_get_private_t) get_private;
      crypt->super.done = (ally_crypt_done_t) done;
      crypt->bits = bits; /* FIXME: Assumes bits > 2. */
      mpz_init (crypt->n);
      mpz_init (crypt->m);
      mpz_init (crypt->p);
      mpz_init (crypt->q);
      mpz_init_set_ui (crypt->e, 65537); /* FIXME: Permit configuration.  Assumes l < 65537. */
      mpz_init (crypt->d);
      mpz_init (crypt->delta);
      mpz_fac_ui (crypt->delta, l);
      mpz_init (crypt->invdelta);
      for (i = 0; i < l; ++i)
        mpz_init (crypt->a[i]);
      gmp_randinit_default (crypt->rnd);
      gmp_randseed_ui (crypt->rnd, time (0)); /* FIXME: Permit configuration. */
    }
  return (ally_crypt_t *) crypt;
}
