/* Copyright (c) 2009, Cornell University.  All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer
       in the documentation and/or other materials provided with the
       distribution.
     * Neither the name of the University nor the names of its
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

/* Package configuration. */
#include "config.h"

/* Standard library. */
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Readline. */
#include <readline/history.h>
#include <readline/readline.h>

/* Ally. */
#include <ally.h>

/* Options. */
static const char opts[] = "";
static const struct option longopts[] = {
  {"version", 0, 0, 0},
  {"help", 0, 0, 0},
  {0, 0, 0, 0},
};

/* Trivial peers set. */
typedef struct trivial_peers_t trivial_peers_t;
struct trivial_peers_t
{
  ally_peers_t super;
  ally_agree_t *agree;
};

/* "Send" messages by immediately handling them. */
int
trivial_send (trivial_peers_t * peers, size_t peer, size_t size,
              const char *message)
{
  peers->agree->message (peers->agree, size, message);
}

/* "Learn" messages by printing them. */
int
trivial_learn (ally_agree_t *agree, size_t size, const char *message)
{
  fwrite (message, 1, size, stdout);
  putchar ('\n');
}

/* Chat program that exercises distributed agreement. */
int
main (int argc, char **argv)
{
  trivial_peers_t peers;
  ally_agree_t *agree;
  ally_crypt_t *crypt;
  char *line;
  int opt;
  int optind;

  /* Process options. */
  while ((opt = getopt_long (argc, argv, opts, longopts, &optind)) != -1)
    switch (opt)
      {
      case 0:
        if (strcmp (longopts[optind].name, "version") == 0)
          {
            puts ("chat - " PACKAGE_TARNAME " " PACKAGE_VERSION "\n");
            puts ("Copyright (C) 2009 Cornell University.\n"
                  "This is free software; see the source for copying conditions.\n"
                  "There is NO WARRANTY, to the extent permitted by law.\n");
            printf ("Written by Robert Burgess.\n");
            return EXIT_SUCCESS;
          }
        else if (strcmp (longopts[optind].name, "help") == 0)
          {
            printf ("Usage: %s [OPTION]... HOST:PORT...\n", argv[0]);
            putchar ('\n');
            puts ("Chat among peers with well-ordered messages.\n");
            puts ("Options:\n"
                  "  --version  Display version information and exit.\n"
                  "  --help     Display usage information and exit.\n");
            printf ("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
            return EXIT_SUCCESS;
          }
      default:
        return EXIT_FAILURE;
      }

  /* Set up the agreement protocol. */
  peers.super.size = 1;
  peers.super.send = (ally_peers_send_t) trivial_send;
  peers.super.done = 0;
  agree =
    ally_agree_fab_paxos ((ally_peers_t *) &peers, (ally_peers_t *) &peers,
                          (ally_peers_t *) &peers,
                          ALLY_AGREE_FAB_PAXOS_ROLE_LEARNER,
                          (ally_agree_learn_t) trivial_learn, 0);
  crypt = ally_crypt_shoup (6, 3, 3, 1024);
  peers.agree = agree;

  /* Interact with the user. */
  for (;;)
    {
      line = readline ("");
      if (!line)
        {
          return EXIT_SUCCESS;
        }
      if (line[0] == '\0')
        {
          free (line);
          continue;
        }
      agree->propose (agree, strlen (line), line);
      add_history (line);
      free (line);
    }

  /* Clean up. */
  agree->done (agree);
  crypt->done (crypt);
}
